/***************************************************************************
 * event.c
 *
 * Copyright (c) 2000 Stefan Hellkvist, Ericsson Radio Systems AB
 * Email: stefan@hellkvist.org
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *****************************************************************************/

/* Oct 2000 - ported to Pocket Viewer by Wittawat Yamwong <wittawat@web.de> */

#include <stdrom.h>

#include "define.h"
#include "libc.h"
/*ce-new2*/
#include "l_define.h"
#include "l_libc.h"
/*ce-new2-end*/

#include "md.h"

#include "point.h"
#include "event.h"
#include "pattern.h"
#include "wizard.h"

/*ce mode-indicators and buttons */
#include "modebut.h"

#include "mod_symb.h"

#include "params.h"


extern BANK bank;

static void paintPatternPart();
static void clearWriteArea();

typedef struct {
    word object;
    bool ink;
    int  ch;
    int  xmid, x, y, width, height;
    Pattern FAR *current;
} PVMerlin;


static PVMerlin Merlin;
static BANK bank;
/*ce static MODE mode;*/
static int timer;

#if 0
/*ce statistical information*/
static int rawNrOfPoints;      /* number of points in original (raw) pattern */
static int rawStartSlope;      /* slope at beginning of raw pattern */
static int rawStartEndVectorX; /* x value of start-to-end-vector (x[last]-x[0]) of raw pattern */
static int rawStartEndVectorV; /* y value of ... (y[last]-y[0]) */
static int rawStartEndSlope;   /* slope of raw start-to-end-vector */

/*ce same statistical values for final (pre-processed) pattern*/
static int nrOfPoints;      /* number of points in final pattern (as used for elastic pattern matching); same as */
static int startSlope;      /* slope at beginning of pattern */
static int startEndVectorX; /* x value of start-to-end-vector (x[last]-x[0]) */
static int startEndVectorV; /* y value of ... (y[last]-y[0]) */
static int startEndSlope;   /* slope of start-to-end-vector */
#endif



static void
findMinMax( Pattern FAR *p, int *xmin_result, int *xmax_result, int *ymin_result, int *ymax_result  )
{
    int minx, maxx, miny, maxy, i;
    Point FAR* pt = &(p->pv[0]);
    minx = maxx = pt->x;
    miny = maxy = pt->y;

    for ( i = 1; i < p->nPoints; i++ )
    {
    pt = &(p->pv[i]);
    if (pt->x > maxx)
        maxx = pt->x;
    if (pt->y > maxy)
        maxy = pt->y;
    if (pt->x < minx)
        minx = pt->x;
    if (pt->y < miny)
        miny = pt->y;
    }
    *xmin_result = minx;
    *xmax_result = maxx;
    *ymin_result = miny;
    *ymax_result = maxy;
}

static void
outputChar( unsigned char c )
{
    Merlin.ch = c;
}


/**
 * setMode
 * Sets the mode to either single or multiple mode
 **/
/*ce
void
setMode( MODE m )
{
    mode = m;
}
*/

static t_char_mode  char_mode  = ALPHA_MODE;
static t_shift_mode shift_mode = LOWERCASE_MODE;

void setCharMode( t_char_mode cm )
{
    char_mode = cm;
}

void setShiftMode( t_shift_mode sm )
{
    shift_mode = sm;
}

t_char_mode getCharMode()
{
    return (shift_mode==SHIFTED_MODE) /* || shift_mode==CAPSLOCK_MODE ) */
            ? ALPHA_MODE  /* SHIFT_MODE always implies ALPHA_MODE */
            : char_mode;
}

t_shift_mode getShiftMode()
{
    return shift_mode;
}

unsigned char changeToNextMode( unsigned char c )
{    
    if ( c == UP_STROKE ) {
        if ( getShiftMode() == LOWERCASE_MODE )
            setShiftMode( SHIFTED_MODE );
        else if ( getShiftMode() == SHIFTED_MODE ) {
            setShiftMode( CAPSLOCK_MODE );
            setCharMode( ALPHA_MODE );
        }
        else /* if CAPSLOCK_MODE */
            setShiftMode( LOWERCASE_MODE );
        modeButtonsDisplayMode(); /* char_mode, shift_mode); */
        return 0;
    } 
    else if ( getShiftMode() == SHIFTED_MODE ) {  /* always fall back to lowercase-alpha from single-char shifted-mode */
        /*ce-new2 */
        setShiftMode( LOWERCASE_MODE );
        if ( c == '.' ) {
            setCharMode( ALPHA_EXT_MODE );
            modeButtonsDisplayMode(); 
            return 0;
        }
        /*ce-new2-end */
        setCharMode( ALPHA_MODE );
        /*ce-new2 movedline setShiftMode( LOWERCASE_MODE ); */
    } 
    else {  /* the extended and numerical modes */
        t_char_mode cm = getCharMode();
        if ( cm == ALPHA_EXT_MODE ) {
            setCharMode( ALPHA_MODE );        
        }
        else if ( cm == NUMERAL_MODE ) {
            if ( c == '.' ) {
                setCharMode( NUMERAL_EXT_MODE );
                modeButtonsDisplayMode(); 
                return 0;
            } else if ( c == NUM_STROKE )
                setCharMode( ALPHA_MODE );        
            /*else leave NUMERAL_MODE */
        }
        else if ( cm == NUMERAL_EXT_MODE ) {
            setCharMode( NUMERAL_MODE );
        }
        else /* ALPHA_MODE */ {
            if ( c == '.' ) {
                setCharMode( ALPHA_EXT_MODE );
                modeButtonsDisplayMode(); 
                return 0;
            } else if ( c == NUM_STROKE )
                setCharMode( NUMERAL_MODE );        
            /*else leave ALPHA_MODE */           
        }
    }
    modeButtonsDisplayMode(); 
    return c;
}

/**
 * restoreBank
 * Restores the bank
 **/
void
restoreBank()
{
    /*setMode( SINGLE_STROKE );*/
    setBank( bank );
}



/**
 * getMode
 * Returns the current mode
 **/
/*ce
MODE
getMode()
{
    return mode;
}
*/

/**
 * storeBank
 * Stores the bank in a safe place to beware of bank robbers perhaps :)
 **/
void
storeBank()
{
    bank = getBank();
}


/*ce: changed + polish */
static unsigned char
my_toupper( int c )
{
    switch (c) {
    case 299:
        return 197;
    case 228:
        return 196;
    case 246:
        return 214;
#ifdef POLISH
    case 0x92:
        return 0x94;
    case 0xEA:
        return 0x7F;
#endif
    default: 
        return toupper( c );
    }
}

#define MAXDOTSIZE 5

/**
 * timedOut
 * Called when a pattern is finished and we should analyze it.
 **/
static void
timedOut()
{
    unsigned char c;
    int xmin, xmax;
    int ymin, ymax;
    int isdot;

    /*ce changed whole logic */
    findMinMax( Merlin.current, &xmin, &xmax , &ymin, &ymax );
    isdot =  (xmax-xmin <= MAXDOTSIZE) && (ymax-ymin <= MAXDOTSIZE);

    if ( isdot ) 
        c = '.'; /* '.' is a hard-wired extended character */
    else {
        int i;
        Pattern FAR *np; /* pre-processed pattern */ 
        np = doallpreprocessing( Merlin.current );
        
        if ( PARAM(PARAM_DEBUG) == P_DBG_PATTERN || PARAM(PARAM_DEBUG) == P_DBG_FULL ) {
            for (i = 0; i < np->nPoints-1; i++) {
                LibGdsLine(  5 + (np->pv[i].x)  , 25 + (np->pv[i].y),
                             5 + (np->pv[i+1].x), 25 + (np->pv[i+1].y));
            }
        }
        
        if ( getCharMode() == NUMERAL_MODE )
            setBank( NUMERAL );
        else if ( getCharMode() == ALPHA_EXT_MODE )
            setBank( ALPHA_EXT );
        else if ( getCharMode() == NUMERAL_EXT_MODE )
            setBank( NUMERAL_EXT );
        else 
            setBank( ALPHA );

        c = recognize( np );
        /*ce-new2 test recogn.speed */
        if ( PARAM(PARAM_DEBUG) == P_DBG_PATTERN || PARAM(PARAM_DEBUG) == P_DBG_FULL ) {
            byte h,m,s=100,s0=0;
            int sec;
            byte p = PARAM(PARAM_DEBUG);
            PARAM(PARAM_DEBUG) = 0;  /* turn debug off for measurement */
            LibGetTime2( &h, &m, &s0 );
            sec = (((int)h)*24 + m)*60;
            do {
                LibGetTime2( &h, &m, &s );
            } while (s==s0); /* wait for the next second to start (to make it a bit more accurate) */
            sec += s;
            for (i=0; i<50; i++)
                c = recognize( np );
            LibGetTime2( &h, &m, &s );
            sec = ((((int)h)*24 + m)*60+s) - sec; /* time diff */
            LibPutMsgDlg("sec: %i", sec );
            PARAM(PARAM_DEBUG) = p;
        }        
        /*ce-new2-end */
       
        if ( getCharMode() == ALPHA_MODE ) {
            #ifdef POLISH            
            c = recognizeMore( np, ALPHA_POLISH ); /* recognize extra polish characters */
            #endif
            if ( getShiftMode() == SHIFTED_MODE || getShiftMode() == CAPSLOCK_MODE ) {
                c = recognizeMore( np, ALPHA_CAPITAL );  /* also try extra capital patterns */
                c = my_toupper(c);
            }
        }
        
        deletePattern( np );
    }
    
    c = changeToNextMode(c);
        
    /*ce not used here if ( mode == MULTIPLE_STROKE )
        restoreBank(); */

    if ( c == -1 ) 
        c=0;
    /* { */
    /*ce old stuff:
     if ( (c == 'i' || c == '1') && xmin > Merlin.xmid - 10 &&
         xmax < Merlin.xmid + 10 )
        c = 'I';        /* Stupid fix to write I */
        outputChar( c );
    /*}*/

    /*ce*/ 
    dropInputPattern();
        
}


void dropInputPattern()
{
    int i;
    Point FAR *pv;
    pv = Merlin.current->pv;
    for (i = 0; i < Merlin.current->nPoints-1; i++, pv++) {
        LibGdsLineCmp(pv[0].x, pv[0].y, pv[1].x, pv[1].y);
        LibGdsDotCmp(pv[0].x, pv[0].y);
    }
    deletePattern( Merlin.current );
    Merlin.current = newPattern( INITIAL_PATTERN_LENGTH, 'a' );
    LibPutDisp();
}

/**
 * penDown
 * Called when the user has put his foot, no, pen down
 **/
void
penDown( int x, int y )
{
    Point po;
    po.x = x;
    po.y = y;
    addPoint( Merlin.current, po );
    Merlin.ch = -1;
    /*if (Merlin.ink && Merlin.current->nPoints <= 1) {
        clearWriteArea();
        PUTDISP;
    }
    */
}


/**
 * penUp
 * Called when the pen is lifted from the pad
 **/
void
penUp( int x, int y )
{
    if ( PARAM(PARAM_DEBUG) == P_DBG_PATTERN || PARAM(PARAM_DEBUG) == P_DBG_FULL ) {
        LibPutMsgDlg("up: %i,%i\rlast: %i,%i", x,y, Merlin.current->pv[Merlin.current->nPoints-1].x, Merlin.current->pv[Merlin.current->nPoints-1].y );
    }
    
    /* if ( mode == SINGLE_STROKE ) */
    timedOut();
    /* else
    timer = 1; */
}


/**
 * penDragged
 * Called when the user drags the pen over the pad
 **/
void
penDragged( int x, int y )
{
    Point po;
    int n;

    if (Merlin.current->nPoints >= MAX_PATTERN_LENGTH)
       return;   /*ce longer patterns are not allowed. Although pattern.c allows dynamic growth of 
                   patterns (which is used if INITIAL length is exceeded) we stop growing at some point, 
                   because it simply would not make much sense (and we will run out of memory anyway) */

    po.x = x;
    po.y = y;
    n = Merlin.current->nPoints;
    addPoint( Merlin.current, po );
    if (Merlin.ink && Merlin.current->nPoints > 1 && Merlin.current->nPoints > n) {
        paintPatternPart();
    }
}

static void clearWriteArea()
{
    LibClrBox(Merlin.x, Merlin.y, Merlin.width, Merlin.height);
    LibBox(Merlin.x, Merlin.y, Merlin.width, Merlin.height, 1);
    /*LibMeshLine(Merlin.xmid, Merlin.y + 1, 1, Merlin.height - 2);*/
}

static void paintPatternPart()
{
    Point FAR *pv;
    int xmin, xmax, ymin, ymax;
    pv = &Merlin.current->pv[Merlin.current->nPoints-2];
    LibGdsLineCmp(pv[0].x, pv[0].y, pv[1].x, pv[1].y);
    LibGdsDotCmp(pv[0].x, pv[0].y);
    if (pv[0].x < pv[1].x ) {
        xmin = pv[0].x;
        xmax = pv[1].x;
    } else {
        xmin = pv[1].x;
        xmax = pv[0].x;
    }
    if (pv[0].y < pv[1].y ) {
        ymin = pv[0].y;
        ymax = pv[1].y;
    } else {
        ymin = pv[1].y;
        ymax = pv[0].y;
    }    
    LibPutDispBox(xmin, ymin, xmax-xmin+1, ymax-ymin+1); 
}


int mlHandleEvent(TCHSTS *tsts)
{
    switch (tsts->act) {
    case ACT_MAKE:
        penDown(tsts->x, tsts->y);
        break;

    case ACT_BREAK_IN:
    case ACT_BREAK:
    case ACT_MOVE_OUT:
        penUp(tsts->x, tsts->y);
        /*LibTchInit(); ce: why the heck whas this line here? */
        break;

    case ACT_MOVE_IN: /* should never happen, but to my suspicion, it happes on 450X Models. !? */
        /*penDown(tsts->x, tsts->y);
        break; */
    case ACT_MOVE:
    case ACT_DOWN_IN:
        penDragged(tsts->x, tsts->y);
        break;
    }
    return Merlin.ch;
}

void mlPaint()
{
    /*int i;
    Point FAR *pv; */

    if (!Merlin.ink) 
        return;
    /*ce clearWriteArea();
    pv = Merlin.current->pv;
    for (i = 0; i < Merlin.current->nPoints-2; i++) {
    LibGdsLine(pv[0].x, pv[0].y, pv[1].x, pv[1].y);
        pv++;
    }
    */
}

static bool firstInit=TRUE;

void mlInit(TCHTBL far *tch, bool ink)
{
    /*ce-kb initWizard(); */
    Merlin.object = tch->obj;
    Merlin.ink    = ink;
    Merlin.x      = tch->x1;
    Merlin.y      = tch->y1;
    Merlin.width  = tch->x2 - tch->x1 + 1;
    Merlin.height = tch->y2 - tch->y1 + 1;
    Merlin.xmid   = (tch->x1 + tch->x2) / 2;
    if (firstInit) {
        initWizard();
        Merlin.current= newPattern( INITIAL_PATTERN_LENGTH, 'a' );
        firstInit = FALSE;
    }
    setBank( ALPHA );
    /* setMode( SINGLE_STROKE ); */
}
