/* 
 * Loading and saving memo items.
 * Christian Ensel   2001/09/27
 * Marcin Woli\'nski 2001/08/12
 */

#include <stdrom.h>
#include "define.h"
#include "libc.h"
#include <loadsave.h>
#include "util.h"   
#include "params.h" 

#define NUM_CATEGORIES 5

static char* getNextPrevPageContents(char *title, t_searchdir d, word *p_fp, int cat);


static int selectcategory(int titleprefixnr) 
{ 
  byte name[16];
  char title[32]; 
  byte rv[NUM_CATEGORIES+1];
  byte sp[NUM_CATEGORIES*16+2]; 
  byte *mp;
  int i;
  mp=sp;
  
  for (i=1;i<=NUM_CATEGORIES;i++)
    {
      LibLblAreaRead(name,4+i); /* Get edited category name... */
      if (*name==0) LibReadMessage(245+i,name); /* or default category name */
      strcpy(mp,name);
      while (*mp!=0) mp++;
      *mp++=0xfe;
      rv[i-1]=i;
    }
  mp--;
  *mp++=0xff; /* end-menu marker */
  *mp=0;
  rv[NUM_CATEGORIES]=0xff;

  /* read title */
  LibReadMessage(154,name);            /* "Category" */
  LibReadMessage(titleprefixnr,title); /* "List/Save/..." */
  strcat(title,": ");
  strcat(title,name);
  return (SelWindowExt2(-1, -1, title , 0, sp, rv ));
}



/* save text in *textbuf in the textfile at *p_fp in categorie *p_cat */
/* fp==NEWFILE -> ask user for categorie and save as a new memo */
/* *p_fp is changed after saving !! */
/* returns success, resp. HALF if user cancels */
int savetext(char *txtbuf, word *p_fp, int *p_cat) {
    return savetextAt( txtbuf, p_fp, p_cat, NEWFILE );
}

/* same as above, but sorts file in at a given place */
int savetextAt(char *txtbuf, word *p_fp, int *p_cat, word prev_fp ) {
  FILE_BUF fb;
  FILE_INFO finf;
  
  if ( *p_cat==-1 ) {  /* category not selected, yet */
      int cat = selectcategory(477);
      if ( cat == 0xff) 
          return HALF; 
      *p_cat = cat;
  }

  fb.fsb_main_entry_ = 3 << 4;	/* Memo main mode code */
  fb.fsb_sub_entry_  = *p_cat;	/* Sub mode code (memo category) */
  fb.fsb_scrt_info_  = CURRENT_FILE_MODE;	
  finf.fp      = *p_fp;
  finf.kind    = FILE_KIND_TEXT; /* Memos are saved as text */
  strcpy(fb.fbuf.text.fsb_text_buf_, txtbuf);
  /*  LibPutMsgDlg("Will write %d:\r%s", cat, fb.fbuf.text.fsb_text_buf_); */
  if ( !LibFileWrite(&fb,&finf) )
      return FALSE;
      
  *p_fp = finf.fp;
  
  if (prev_fp!=NEWFILE) {
      LibFileExch( &fb, &finf, prev_fp);
  }
  return TRUE;
}


/* load new text via user query, 
   (if title!=NULL it only displays memos with the given title)
   returns whether new file was loaded, 
   returns HALF if user cancels
   if yes: writes loaded file-pointer to *p_fp */
int loadtext(char *txtbuf, char* title, word *p_fp, int *p_cat) {
  char dialogtitle[25];
  int cat;
  byte sel;
  bool fewItems=TRUE;
  word fp;
  word fp_test;
  char* pfiletxt=NULL;
  byte sp[(MAX_TITLE_LEN+1)*15+3], *from, *to; /* Menu (list) text */ 
  byte rv[17]; /* Return values */
  word fps[17];   /* for saving file pointers */
  int i;

    if (title) {
        strcpy(dialogtitle,"Goto-Page: "); /* 11 chars */
        strncat(dialogtitle, title, 11);   /* +11 chars */
        strcat(dialogtitle, "..");         /* +2 = 24 chars */
    }    
    else 
        strcpy(dialogtitle,"Load-Memo");

  do {
      if ( *p_cat==-1 ) {  /* category not selected, yet */
          cat = selectcategory(49);
          if (cat == 0xff) 
              return HALF; 
      }
      else 
          cat = *p_cat;
      /* Now cat is the category to load memos from */
    
      /* Start building memo list */
      do {
          if (pfiletxt==NULL)
              fp = 0xffff;
          to=sp; 
          i=0;  
          pfiletxt = getNextPrevPageContents( title, SEARCH_NEXT_PAGE, &fp, cat);  /*LibFileFindNext(&fb,&finf,0x00); */
          if (pfiletxt==NULL) {  /* if last round ended exactly at the last memo, then pfiletxt was !=NULL, but nevertheless we have to restart loading with the first memo */
              fp = 0xffff;
              pfiletxt = getNextPrevPageContents( title, SEARCH_NEXT_PAGE, &fp, cat);  /*LibFileFindNext(&fb,&finf,0x00); */
              if (pfiletxt==NULL) {
                  LibPutMsgDlg("No memos in\rcategory %i",cat);
                  sel = 0xff;
                  break; /*return HALF;*/
              }
          }
          /* enter title as menu item */
          do {
            int charcnt=0;
            fps[i] = fp;
            from = pfiletxt;    
            if (title) { /* show only subtitle */
                while ( !IS_TITLE_ENDCHAR(*from) )
                    from++;
                if (*from == '-' || *from == ':') {  /* '(' remains in subtitle */
                    from++;
                    while ( *from==' ' )  /* skip space */
                      from++;
                }
                if (*from==0 || *from=='\r') { /* no title or no subtitle */
                    if (fewItems)
                        sprintf(to,"[unnamed page #%i]", i+1 );
                    else
                        strcpy(to,"[unnamed page]");
                    to += strlen(to); /* set to at the end again */
                }
            }
            while (*from!=0 && *from!='\r' && charcnt<=MAX_TITLE_LEN) { /* limit length */
                charcnt++;
                *to++=*from++;
            }
            *to++=0xfe;  /* end-item marker */
            rv[i]=i;
            i++;
          } while ( i<14 && 
                    (pfiletxt=getNextPrevPageContents( title, SEARCH_NEXT_PAGE, &fp, cat)) /* (more=LibFileFindNext(&fb,&finf,0x00)) /* !!! */ 
                    );
          fp_test = fp;
          if ( fewItems==FALSE || 
               (i==14 && getNextPrevPageContents( title, SEARCH_NEXT_PAGE, &fp_test, cat)) /*LibFileFindNext(&fb,&finf,0x00)*/ ) { 
              /* if this is not the first round or we have a full item-list and there are more files */
              fewItems = FALSE;
              strcpy(to, "[more ...]");
              to += strlen(to);
              rv[i]=15;
              i++;
          }
          else 
              to--;
          *to++=0xff; /* end-menu marker  */
          *to=0;
          rv[i]=0xff;
        
          sel = SelWindowExt2(-1, -1, dialogtitle , 0, sp, rv );
        
          if (sel == 0xff) { /* esc from file selection */
              if (*p_cat==-1)
                  break; /* return to categoty selection */
              else 
                  return HALF; /* or end file selection completely */
          }
      } while ( sel ==15 );
  
  } while (sel==0xff); /* while user <esc>apes out of file selection, go back to category selection */
  
  if ( !loadpage(txtbuf, NULL, SEARCH_THIS_PAGE, &fps[sel], cat) )
      return FALSE;
     
  *p_fp = fps[sel];
  *p_cat = cat;
  return TRUE;
}

/* load next/prev page after *p_fp with a certain title (or title==null to load any memo)
   (search for first page, if *p_pf==0xffff; last page==0xfffe */
/* txtbuf and title are allowed to point to the same buffer -> title will be overridden by text */
/* *p_fp is set to new page */
/* returns TRUE for success, FALSE otherwise */
int loadpage(char *txtbuf, char *title, t_searchdir d, word *p_fp, int cat)
{
    char *res = getNextPrevPageContents( title, d, p_fp, cat);
    if (res==NULL)
        return FALSE;
    strcpy(txtbuf, res);
    return TRUE;
}


/*ce internal load function */
/* load next/prev page (according to given 'dir'ection) after a given memo
   at '*p_fp' with a certain title (or title==null to load any memo)
   (search for first page, if *p_pf==0xffff; last page==0xfffe)
   in the category 'cat' */
/* on success, '*p_fp' is changed to the new file-pointer */
/* returns NULL if no page/memo was found, otherwise a pointer to the 
   memo-contents */
static char* getNextPrevPageContents(char *title, t_searchdir d, word *p_fp, int cat)
{
    int found;
    int fileok;
    int l;
    FILE_BUF fb;
    FILE_INFO finf;
    l = (title ? strlen(title) : 0);
    fb.fsb_main_entry_ = 3 << 4;	/* Memo main mode code */
    fb.fsb_sub_entry_  = cat;	/* Sub mode code (memo category) */
    fb.fsb_scrt_info_  = CURRENT_FILE_MODE;
    finf.fp      = *p_fp;
    finf.kind    = FILE_KIND_TEXT; /* Memos are saved as text */

    /* if we search for first or last page we ignore value of parameter d */
    if ( finf.fp == 0xffff)
        d = SEARCH_NEXT_PAGE;
    else if ( finf.fp == 0xfffe )
        d = SEARCH_PREV_PAGE;
    
    do { /* iterate over files until right title is found */
        if ( d == SEARCH_NEXT_PAGE || finf.fp == 0xfffe ) /* search forward or for very last item */
            fileok = LibFileFindNext(&fb,&finf,0x00);
        else if ( d == SEARCH_PREV_PAGE )
            fileok = LibFileFindPrev(&fb,&finf,0x00);
        if (fileok || d == SEARCH_THIS_PAGE )
            fileok = LibFileRead(&fb,&finf);
        if (fileok) {
            if ( l==0 )
                found=TRUE;
            else {
                found = !strncmp(title, fb.fbuf.text.fsb_text_buf_, min(l,MAX_TITLE_LEN) ); /* check title prefix */
                if (found && l<MAX_TITLE_LEN ) { /* check whether loaded title has same length */
                                                 /* (-> titles are equal, test whether searchtitle is not just a prefix of the loaded one - only for titles longer than the maximum, this is not checked ) */
                    char *tend = fb.fbuf.text.fsb_text_buf_+l;
                    while ( *tend==' ' ) /*skip spaces*/
                      tend++;
                    found = IS_TITLE_ENDCHAR(*tend);
                }
                if (d == SEARCH_THIS_PAGE && !found)
                    return NULL;
            }
        } 
    } while ( fileok && !found );
    
    if (!fileok || !found)
        return NULL;
        
    *p_fp = finf.fp;
    return fb.fbuf.text.fsb_text_buf_;    
}

int deletefile(word fp) {
    FILE_BUF unusedfb;
    FILE_INFO finf;
    finf.fp      = fp;
    finf.kind    = FILE_KIND_TEXT; 
    return LibFileRemove( &unusedfb, &finf );
}